/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2011 Klaus Spanderen

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.

 compile:
g++ -DNTHREADS=8 -I/home/spanderen/workspace/QuantLib -I/usr/local/include/boost-1_38 -O3 -o vpp vpp5.cpp -L/home/spanderen/workspace/QuantLib/ql/.libs -lQuantLib -lboost_thread-gcc42-mt
*/

#include <ql/settings.hpp>
#include <ql/math/functional.hpp>
#include <ql/time/daycounters/actualactual.hpp>
#include <ql/math/randomnumbers/rngtraits.hpp>
#include <ql/math/randomnumbers/sobolrsg.hpp>
#include <ql/math/statistics/generalstatistics.hpp>
#include <ql/models/marketmodels/browniangenerators/sobolbrowniangenerator.hpp>
#include <ql/instruments/vanillaswingoption.hpp>
#include <ql/methods/montecarlo/multipathgenerator.hpp>
#include <ql/experimental/processes/klugeextouprocess.hpp>
#include <ql/experimental/processes/extouwithjumpsprocess.hpp>
#include <ql/experimental/processes/extendedornsteinuhlenbeckprocess.hpp>
#include <ql/experimental/finitedifferences/uniform1dmesher.hpp>
#include <ql/experimental/finitedifferences/fdmvppstepcondition.hpp>
#include <ql/experimental/finitedifferences/fdmmeshercomposite.hpp>
#include <ql/experimental/finitedifferences/fdmlinearoplayout.hpp>
#include <ql/experimental/finitedifferences/fdminnervaluecalculator.hpp>
#include <ql/experimental/finitedifferences/fdsimpleklugeextouvppengine.hpp>

#include <boost/shared_array.hpp>
#include <boost/thread/thread.hpp>
#include <boost/thread/mutex.hpp>

#include <fstream>

#ifndef NTHREADS
#define NTHREADS 1
#endif

using namespace QuantLib;

namespace {

    // model definition
    const Real beta         = 200;
    const Real eta          = 1.0/0.4;
    const Real lambda       = 4.0;
    const Real alpha        = 7.0;
    const Real volatility_x = 1.4;
    const Real kappa        = 4.45;
    const Real volatility_u = std::sqrt(1.3);
    const Real rho          = 0.7;
    const Rate irRate       = 0.00;

    // vpp definition
    const Size maturityInWeeks = 26;
    const Real efficiencies[] 
        = { 0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.60 }; // MWh power per MWh heat
    const Real pMin             = 8;    // MWh
    const Real pMax             = 40;   // MWh
    const Real startUpFuel      = 20;   // MWh
    const Real startUpFixedCost = 300;  // EUR
    const Real carbonPrice      = 3.0;  // EUR per MWh heat (from gas)
    const Size minOn            = 6.0;  // hour
    const Size minOff           = 2.0;  // hour
    const Size nStates          = 2u*minOn + minOff;

    const Size nTrails = 1024;

    typedef PseudoRandom::rsg_type rsg_type;
    typedef MultiPathGenerator<rsg_type>::sample_type sample_type;
    typedef FdSimpleKlugeExtOUVPPEngine::Shape Shape;


    class PathGasPrice : public FdmInnerValueCalculator {
      public:
        typedef FdSimpleKlugeExtOUVPPEngine::Shape Shape;

        PathGasPrice(
            const MultiPathGenerator<PseudoRandom>::sample_type::value_type& path,
            const boost::shared_ptr<Shape>& shape)
        : path_(path),
          shape_(shape) {}
        Real innerValue(const FdmLinearOpIterator&, Time t) {
            const Size i = (Size) (t*365*24);
            const Real f = std::lower_bound(shape_->begin(), shape_->end(),
                std::pair<Time, Real>(t-std::sqrt(QL_EPSILON), 0.0))->second;

            return std::exp(path_[2][i] + f);
        }
        Real avgInnerValue(const FdmLinearOpIterator& iter, Time t) {
            return innerValue(iter, t);
        }
      private:
        const MultiPathGenerator<PseudoRandom>::sample_type::value_type& path_;
        const boost::shared_ptr<Shape> shape_;
    };

    class PathSparkSpreadPrice : public FdmInnerValueCalculator {
      public:
        typedef FdSimpleKlugeExtOUVPPEngine::Shape Shape;

        PathSparkSpreadPrice(
            Real heatRate,
            const MultiPathGenerator<PseudoRandom>::sample_type::value_type& path,
            const boost::shared_ptr<Shape>& gasShape,
            const boost::shared_ptr<Shape>& powerShape)
        : heatRate_(heatRate),
          path_(path),
          gasShape_(gasShape),
          powerShape_(powerShape) {}

        Real innerValue(const FdmLinearOpIterator&, Time t) {
            const Size i = (Size) (t*365*24);
            const Real f = std::lower_bound(
                powerShape_->begin(), powerShape_->end(),
                std::pair<Time, Real>(t-std::sqrt(QL_EPSILON), 0.0))->second;
            const Real g = std::lower_bound(
                gasShape_->begin(),gasShape_->end(),
                std::pair<Time, Real>(t-std::sqrt(QL_EPSILON), 0.0))->second;
            
            return std::exp(f + path_[0][i]+path_[1][i])
                    - heatRate_*std::exp(g + path_[2][i]);
        }
        Real avgInnerValue(const FdmLinearOpIterator& iter, Time t) {
            return innerValue(iter, t);
        }
      private:
        const Real heatRate_;
        const MultiPathGenerator<PseudoRandom>::sample_type::value_type& path_;
        const boost::shared_ptr<Shape> gasShape_;
        const boost::shared_ptr<Shape> powerShape_;
    };


    class SobolQRsg {
    public:
        typedef SobolRsg::sample_type sample_type;

        explicit SobolQRsg(Size dimensionality, unsigned long seed = 1234)
        : rsg_(dimensionality, 1234, SobolRsg::JoeKuoD7),
          seq_(std::vector<Real>(dimensionality), 1.0),
          offsets_(dimensionality) {
            MersenneTwisterUniformRng rng(seed);
            for (Size i=0; i < dimensionality; ++i) {
                offsets_[i] = rng.nextReal();
            }
        }

        const SobolQRsg::sample_type& nextSequence() const {
            seq_ = rsg_.nextSequence();
            for (Size i=0; i < seq_.value.size(); ++i) {
                seq_.value[i]+=offsets_[i];
                if (seq_.value[i] >= 1.0) {
                    seq_.value[i] -=1.0;
                }
            }
            return seq_;
        }

        const sample_type& lastSequence() const { return seq_; }
        Size dimension() const { return rsg_.dimension(); }
    private:
        SobolRsg rsg_;
        mutable sample_type seq_;
        std::vector<Real> offsets_;
    };

    class Worker {
    public:
        Worker(Real heatRate,
               const boost::shared_ptr<FdmMesher>& mesher,
               const boost::shared_ptr<Shape>& gasShape,
               const boost::shared_ptr<Shape>& powerShape,
               const boost::shared_ptr<Shape>& gasZeroVolShape,
               const boost::shared_ptr<Shape>& powerZeroVolShape,
               const TimeGrid& grid,
               const boost::shared_ptr<KlugeExtOUProcess>& klugeOUProcess)
        : localIndex_(index++),
          heatRate_(heatRate),
          results_(new Real[6]),
          mesher_(mesher),
          gasShape_(gasShape),
          powerShape_(powerShape),
          gasZeroVolShape_(gasZeroVolShape),
          powerZeroVolShape_(powerZeroVolShape),
          grid_(grid),
          klugeOUProcess_(klugeOUProcess) {
        }

        void operator()() const {
            boost::mutex::scoped_lock lock(mutex[localIndex_ % NTHREADS]);

            GeneralStatistics npv;
            rsg_type rsg = PseudoRandom::make_sequence_generator(
                klugeOUProcess_->factors()*(grid_.size()-1), 1234ul);

            const boost::shared_ptr<MultiPathGenerator<rsg_type> > vanillaGen(
                new MultiPathGenerator<rsg_type>(klugeOUProcess_, grid_, rsg, false));

            for (Size j=0; j < nTrails; ++j) {
                const sample_type& path = vanillaGen->next();
                npv.add(pathPricer(path));
            }
            results_[0] = npv.mean();
            results_[1] = npv.errorEstimate();

            npv.reset();
            for (Size j=0; j < nTrails; j+=2) {
                npv.add(0.5*( pathPricer(vanillaGen->next())
                             +pathPricer(vanillaGen->antithetic())));
            }
            results_[2] = npv.mean();
            results_[3] = npv.errorEstimate();


            npv.reset();

            const Size factors = klugeOUProcess_->factors();
            const Size steps   = grid_.size()-1;
            const Size N       = factors*steps;
            MersenneTwisterUniformRng rng(123456ul);

            for (Size u=0; u < 8; ++u) {
                SobolQRsg sobol(N, rng.nextInt32());
                InverseCumulativeRsg<SobolQRsg,InverseCumulativeNormal> sobolGen(sobol);

                std::vector<std::vector<Real> > variates(
                                         factors*steps, std::vector<Real>(nTrails));
                for (Size j=0; j < nTrails; ++j) {
                    const SobolQRsg::sample_type& sample = sobolGen.nextSequence();
                    for (Size k=0; k < N; ++k) {
                        variates[k][j] = sample.value[k];
                    }
                }
                
                std::vector<std::vector<Real> > bb = SobolBrownianGenerator(
                    klugeOUProcess_->factors(), grid_.size()-1,
                    SobolBrownianGenerator::Diagonal).transform(variates);

                GeneralStatistics localNpv;
                
                for (Size j=0; j < nTrails; ++j) {
                    Array x0 = klugeOUProcess_->initialValues();
                    sample_type path(MultiPath(x0.size(), grid_), 1.0);

                    for (Size k=0; k < steps; ++k) {
                        Array dw(klugeOUProcess_->factors());
                        for (Size l=0; l < dw.size(); ++l) {
                            dw[l] = bb[l][j*steps+k];
                        }
                        x0 = klugeOUProcess_->evolve(grid_.at(k), x0, grid_.dt(k), dw);
                        
                        for (Size l=0; l < x0.size(); ++l) {
                            path.value[l][k] = x0[l];
                        }
                    }
                    
                    localNpv.add(pathPricer(path));
                }
                npv.add(localNpv.mean());
            }
            results_[4] = npv.mean();
            results_[5] = npv.standardDeviation();
        }
 
        Real mcNPV()     const { return results_[0]; }
        Real mcError()   const { return results_[1]; }
        Real antiNPV()   const { return results_[2]; }
        Real antiError() const { return results_[3]; }
        Real quasiNPV()  const { return results_[4]; }
        Real quasiError()  const { return results_[5]; }

        Real heatRate()  const { return heatRate_; }

    private:
        Real pathPricer(const sample_type& path) const {
            FdmVPPStepCondition stepCondition(
                heatRate_, pMin, pMax, minOn, minOff,
                startUpFuel, startUpFixedCost, carbonPrice, 0u, mesher_,
                boost::shared_ptr<FdmInnerValueCalculator>(
                    new PathGasPrice(path.value, gasShape_)),
                boost::shared_ptr<FdmInnerValueCalculator>(
                    new PathSparkSpreadPrice(heatRate_, path.value,
                                             gasShape_, powerShape_)));

            Array state(nStates, 0.0);
            for (Size j=grid_.size(); j > 0; --j) {
                stepCondition.applyTo(state, grid_.at(j-1));
            } 

            return state.back();
        }

        Size localIndex_;
        const Real heatRate_;
        const boost::shared_array<Real> results_;
        const boost::shared_ptr<FdmMesher> mesher_;
        const boost::shared_ptr<Shape> gasShape_;
        const boost::shared_ptr<Shape> powerShape_;
        const boost::shared_ptr<Shape> gasZeroVolShape_;
        const boost::shared_ptr<Shape> powerZeroVolShape_;
        const TimeGrid& grid_;
        const boost::shared_ptr<KlugeExtOUProcess> klugeOUProcess_;

        static Size index;
        static boost::shared_array<boost::mutex> mutex;
    };

    Size Worker::index = 0;
    boost::shared_array<boost::mutex> Worker::mutex(new boost::mutex[NTHREADS+1]);
}



int main(int argc, char *argv[]) {

    const DayCounter dc = ActualActual();
    const Date today = Date(28, Sep, 2010);
    Settings::instance().evaluationDate() = today;

    const Size maturityInDays  = 7*maturityInWeeks;
    const Size maturityInHours = 24*maturityInDays;
    Date maturityDate = today + Period(maturityInWeeks, Weeks);

    boost::shared_ptr<Shape> gasShape(new Shape(maturityInHours));
    boost::shared_ptr<Shape> powerShape(new Shape(maturityInHours));
    boost::shared_ptr<Shape> gasZeroVolShape(new Shape(maturityInHours));
    boost::shared_ptr<Shape> powerZeroVolShape(new Shape(maturityInHours));

    const boost::shared_ptr<SwingExercise> exercise(
        new SwingExercise(today, today+maturityInDays-1, 3600u));
    const std::vector<Time> exerciseTimes = exercise->exerciseTimes(dc, today);

    // read shape from file
    // format is Date, 0h price, 1h price, 2h price, ... 23h price
    // example:4-Oct-2010 36.73 32.09 27.32 23.22 25.71 35.60 47.32 ...
    std::ifstream eex("EEX_2010-2013.txt");
    QL_REQUIRE(eex.is_open(), "can't find price file EEX_2010-2013.txt");
    
    std::ifstream ttf("TTF_2010-2013.txt");
    QL_REQUIRE(ttf.is_open(), "can't find price file TTF_2010-2013.txt");

    Real p;
    for (Size i=0; i < maturityInDays; ++i) {
        std::string str;
        eex >> str;
        ttf >> str;
        
        for (Size j=0; j < 24; ++j) {
            const Size ij = i*24+j;
            const Time t = exerciseTimes[ij]; 
            
            eex >> p;
            (*powerShape)[ij] = Shape::value_type(t, 
                 std::log(p)-square<Real>()(volatility_x)
                             /(4*alpha)*(1-std::exp(-2*alpha*t))
                 -lambda/beta*std::log((eta-std::exp(-beta*t))/(eta-1.0))); 
            (*powerZeroVolShape)[ij] = Shape::value_type(t, std::log(p));

            ttf >> p;
            (*gasShape)[ij] = Shape::value_type(t, 
                 std::log(p)-square<Real>()(volatility_u)
                             /(4*kappa)*(1-std::exp(-2*kappa*t)));
            (*gasZeroVolShape)[ij] = Shape::value_type(t, std::log(p));
        }
    }
    eex.close();
    ttf.close();
 
    const boost::shared_ptr<FdmMesher> oneDimMesher(
        new FdmMesherComposite(boost::shared_ptr<FdmLinearOpLayout>(
            new FdmLinearOpLayout(std::vector<Size>(1, nStates))),
                std::vector<boost::shared_ptr<Fdm1dMesher> >(1,
                    boost::shared_ptr<Fdm1dMesher>(
                        new Uniform1dMesher(0.0, 1.0, nStates)))));

    TimeGrid grid(exerciseTimes.begin(), exerciseTimes.end(), 1);

    Array x0(2);
    x0[0] = 0.0; x0[1] = 0.0;

    boost::shared_ptr<ExtendedOrnsteinUhlenbeckProcess> ouProcess(
        new ExtendedOrnsteinUhlenbeckProcess(alpha, volatility_x, x0[0],
                                             constant<Real, Real>(x0[0])));
    boost::shared_ptr<ExtOUWithJumpsProcess> lnPowerProcess(
           new ExtOUWithJumpsProcess(ouProcess, x0[1], beta, lambda, eta));
    
    Real u=0.0;
    boost::shared_ptr<ExtendedOrnsteinUhlenbeckProcess> lnGasProcess(
        new ExtendedOrnsteinUhlenbeckProcess(kappa, volatility_u, u,
                                             constant<Real, Real>(u)));
    
    const boost::shared_ptr<KlugeExtOUProcess> klugeOUProcess(
        new KlugeExtOUProcess(rho, lnPowerProcess, lnGasProcess));

    const Size nWorker = sizeof(efficiencies)/sizeof(Real);
    std::vector<boost::shared_ptr<Worker> > workers(nWorker);
    boost::thread_group threadGroup;

    for (Size i=0; i < nWorker; ++i) {
        const Real heatRate = 1.0/efficiencies[i];

        workers[i] = boost::shared_ptr<Worker>(
            new Worker(heatRate, oneDimMesher, 
                       gasShape, powerShape, 
                       gasZeroVolShape, powerZeroVolShape,
                       grid, klugeOUProcess));
        threadGroup.add_thread(new boost::thread(*workers[i].get()));
    }

    threadGroup.join_all();
    for (Size i=0; i < nWorker; ++i) {
        printf("efficiency: %2.0f%% "
               " MC %6.0f(%4.0f)  Antithetic %6.0f(%4.0f)  Quasi %6.0f(%4.0f)\n",
               1/workers[i]->heatRate()*100, 
               workers[i]->mcNPV(), workers[i]->mcError(),
               workers[i]->antiNPV(), workers[i]->antiError(),
               workers[i]->quasiNPV(), workers[i]->quasiError());
    }    

    return 0;
}
